@@ -6,7 +6,8 @@ import logging
 from typing import Dict, Collection, Tuple, Union, List
 from typing_extensions import Literal
 
-from fdsreader.utils import Dimension, Quantity, Mesh, Extent
+from fdsreader.fds_classes import Mesh
+from fdsreader.utils import Dimension, Quantity, Extent
 from fdsreader import settings
 import fdsreader.utils.fortran_data as fdtype
 
@@ -51,6 +52,8 @@ class SubSlice:
         """
         shape = self.dimension.shape(cell_centered=self.cell_centered)
         if self.orientation != 0:
+            if len(shape) < 2:
+                return shape[0], 1
             return shape[0], shape[1]
         return shape
 
@@ -79,9 +82,8 @@ class SubSlice:
         n = self.dimension.size(cell_centered=False)
         dtype_data = fdtype.combine(fdtype.FLOAT, fdtype.new((('f', n),)))
 
-        load_times = False
-        if self.n_t == -1:
-            load_times = True
+        load_times = self.n_t == -1
+        if load_times:
             self._parent_slice.n_t = (os.stat(file_path).st_size - self._offset) // dtype_data.itemsize
             self._parent_slice.times = np.empty(self.n_t)
 
@@ -134,8 +136,8 @@ class SubSlice:
         """
         if hasattr(self, "_data"):
             del self._data
-        if hasattr(self, "_vector_data"):
             del self._vector_data
+            self._vector_data = dict()
 
     def __repr__(self):
         return f"SubSlice(shape={self.shape}, mesh={self.mesh.id}, extent={self.extent})"
@@ -143,11 +145,15 @@ class SubSlice:
 
 class Slice(np.lib.mixins.NDArrayOperatorsMixin):
     """Slice file data container including metadata. Consists of multiple subslices, one for each
-        mesh the slice cuts through.
+        mesh the slice cuts through. In case a slice cuts right through the border of two meshes, the generated data
+        would be duplicated. For edge-centered slices a random of both generated slices will be discarded as the data
+        is completely identical. For cell-centered slices the data of both slices will be saved as :class:`SubSlice`s,
+        therefore it might seem as if all slices would be duplicated, in reality however the slices might contain
+        different data.
 
     :ivar cell_centered: Indicates whether centered positioning for data is used.
     :ivar quantity: Quantity object containing information about the quantity calculated for this
-        slice with the corresponding label and unit.
+        slice with the corresponding short_name and unit.
     :ivar times: Numpy array containing all times for which data has been recorded.
     :ivar n_t: Total number of time steps for which output data has been written.
     :ivar orientation: Orientation [1,2,3] of the slice in case it is 2D, 0 otherwise.
@@ -178,12 +184,9 @@ class Slice(np.lib.mixins.NDArrayOperatorsMixin):
 
         for mesh_data in multimesh_data:
             if mesh_data["mesh"] not in self._subslices:
-                self.quantity = Quantity(mesh_data["quantity"], mesh_data["label"],
-                                         mesh_data["unit"])
-                self._subslices[mesh_data["mesh"]] = SubSlice(self, mesh_data["filename"],
-                                                              mesh_data["dimension"],
-                                                              mesh_data["extent"],
-                                                              mesh_data["mesh"])
+                self.quantity = Quantity(mesh_data["quantity"], mesh_data["short_name"], mesh_data["unit"])
+                self._subslices[mesh_data["mesh"]] = SubSlice(self, mesh_data["filename"], mesh_data["dimension"],
+                                                              mesh_data["extent"], mesh_data["mesh"])
 
             if "-VELOCITY" in mesh_data["quantity"]:
                 vector_temp[mesh_data["mesh"]][mesh_data["quantity"]] = mesh_data["filename"]
@@ -470,9 +473,9 @@ class Slice(np.lib.mixins.NDArrayOperatorsMixin):
 
     def __repr__(self):
         if self.type == '3D':  # 3D-Slice
-            return f"Slice([3D] cell_centered={self.cell_centered}, extent={self.extent})"
+            return f"Slice([3D] quantity={self.quantity}, cell_centered={self.cell_centered}, extent={self.extent})"
         else:  # 2D-Slice
-            return f"Slice([2D] cell_centered={self.cell_centered}, extent={self.extent}, " \
+            return f"Slice([2D] quantity={self.quantity}, cell_centered={self.cell_centered}, extent={self.extent}, " \
                    f"extent_dirs={self.extent_dirs}, orientation={self.orientation})"
 
 # __array_function__ implementations
